// Fuzzy Logic Shuttle Controller
// Written by David J. Stein, Esq. See "License.html" for terms and conditions of use.
//
// This module includes the code implementing the fuzzy variable control system underlying this control application.

#region Using declarations

using System;
using System.Collections;

#endregion

namespace Fuzzy_Logic_Shuttle_Controller_Namespace {

	public abstract class FuzzyVariable {
		// This class represents a fuzzy variable.  This is an abstract class - should be overridden by fuzzy variables with particular membership functions.

		#region Variables and initialization code

		public double dMin, dMax;  // the bounds of the fuzzy control variable
		public string strName;         // the name of the fuzzy control variable

		public FuzzyVariable() {
			dMin = dMax = 0;
		}

		#endregion

		#region Core functions

		public abstract double MembershipFunction(double dInput);

		#endregion

	}

	public class FuzzyVariable_Triangular : FuzzyVariable {
		// This subclass of FuzzyVariable represents a fuzzy variable with a triangular member function. Triangle is defined by dMin, dMax, and amplitude.

		#region Variables and initialization code

		private double dAmplitude;

		public FuzzyVariable_Triangular(string strName, double dMin, double dMax, double dAmplitude) {
			this.strName = strName;
			if (dMin >= dMax) {
				dMin = dMax = dAmplitude = 0;
			}
			else {
				this.dMin = dMin;
				this.dMax = dMax;
				this.dAmplitude = dAmplitude;
			}
		}

		#endregion
		
		#region Core functions

		public override double MembershipFunction(double dInput) {
			// This is the triangular membership function.
			if ((dInput <= dMin) || (dInput >= dMax))
				return 0.0;
			else
				return dAmplitude * (1.0f - Math.Abs((dMax - dMin) / 2.0f + dMin - dInput) / (dMax - dMin) * 2.0);
		}

		#endregion

	}

	public class FuzzyVariableMembership {
		// This class represents the membership for a certain fuzzy variable.

		#region Variables and initialization code

		public FuzzyVariable variable; 
		public double dMembership;
		public FuzzyVariableMembership(FuzzyVariable variable, double dMembership) {
			this.variable = variable;
			this.dMembership = dMembership;
		}

		#endregion

	}

	public class FuzzyVariableSet {
		// This class represents a fuzzy variable set, and includes many functions for working with fuzzy variables.

		#region Variables and initialization code

		private ArrayList aFuzzyVariables;  // array of FuzzyVariable objects
		private FuzzyVariable[,] aFuzzyRules;  // array of FuzzyVariable rules
		public double dMin, dMax, dIntegrationInterval;

		public FuzzyVariableSet(double dIntegrationInterval) {
			aFuzzyVariables = new ArrayList();
			aFuzzyRules = null;
			dMin = 0;
			dMax = 0;
			this.dIntegrationInterval = dIntegrationInterval;
		}

		#endregion

		#region Core functions

		public void AddVariable(FuzzyVariable newVariable) {
			// This function adds a fuzzy variable to the fuzzy variable set and expands the size of the aFuzzyRules array. All rules are initially set to null (no rule.)
			dMin = Math.Min(dMin, newVariable.dMin);
			dMax = Math.Max(dMax, newVariable.dMax);
			aFuzzyVariables.Add(newVariable);
			aFuzzyRules = new FuzzyVariable[aFuzzyVariables.Count, aFuzzyVariables.Count];
			for (int i = 0; i < aFuzzyVariables.Count; i++) {
				for (int j = 0; j < aFuzzyVariables.Count; j++) {
					aFuzzyRules[i,j] = null;
				}
			}
		}

		public void AddRule(FuzzyVariable variable1, FuzzyVariable variable2, FuzzyVariable result) {
			// This function adds a result fuzzy variable to the aFuzzyRules array.
			int index1 = aFuzzyVariables.IndexOf(variable1);
			int index2 = aFuzzyVariables.IndexOf(variable2);
			if ((index1 != -1) && (index2 != -1))
				aFuzzyRules[index1, index2] = result;
		}

		public ArrayList FindVariables(double dInput) {
			// This function finds all variables associated with an input value. It returns a list of FuzzyVariableMembership objects (e.g., "0.2 = fuzzy variable ZERO with 0.6 weight and fuzzy variable POS_SMALL with 0.7 weight.")
			ArrayList aList = new ArrayList();
			foreach (FuzzyVariable variable in aFuzzyVariables) {
				double dMembership = variable.MembershipFunction(dInput);
				if (dMembership > 0)
					aList.Add(new FuzzyVariableMembership(variable, dMembership));
			}
			return aList;
		}

		public string ReportCurrentVariable(double dInput) {
			// this function finds the strongest fuzzy variable associated with this value and returns a strong with the name of the variable and its membership.
			double dBestMembership = 0.0;
			string strReport = "None";
			foreach (FuzzyVariable variable in aFuzzyVariables) {
				double dMembership = variable.MembershipFunction(dInput);
				if (dMembership > dBestMembership) {
					dBestMembership = dMembership;
					strReport = variable.strName;
				}
			}
			return strReport + " (" + Math.Round((decimal) dBestMembership, 8) + ")";
		}

		public FuzzyVariableMembership FindFuzzyRule(FuzzyVariableMembership variable1, FuzzyVariableMembership variable2) {
			// This function finds the fuzzy rule for this set of input variables. It returns a FuzzyVariableMembership object - but the contained membership function is the min of the memberships of the two passed-in variables.
			int index1 = aFuzzyVariables.IndexOf(variable1.variable);
			int index2 = aFuzzyVariables.IndexOf(variable2.variable);
			if ((index1 == -1) || (index2 == -1))
				return null;
			else
				return new FuzzyVariableMembership(aFuzzyRules[index1, index2], Math.Min(variable1.dMembership, variable2.dMembership));
		}

		public double FuzzyResult(double dInput1, double dInput2) {
			// This is the main fuzzy variable function - it accepts two input values and outputs the result value, based on fuzzy inference and de-fuzzification.
			// Step 1: Find all fuzzy variables and memberships for each input value.
			ArrayList dInput1Variables = FindVariables(dInput1);
			ArrayList dInput2Variables = FindVariables(dInput2);
			// Step 2: Find all fuzzy output values for each combination of dInput1Variables and dInput2Variables, using the rules table for this fuzzy variable set.
			ArrayList dOutputVariables = new ArrayList();
			foreach (FuzzyVariableMembership dVariable1 in dInput1Variables) {
				foreach (FuzzyVariableMembership dVariable2 in dInput2Variables) {
					dOutputVariables.Add(FindFuzzyRule(dVariable1, dVariable2));
				}
			}
			// Step 3: Calculate the result of fuzzy variables across all memberships.
			double[] dResults = new double[(int) ((dMax - dMin) / dIntegrationInterval) + 1];
			int i = 0;
			for (double dIncrement = dMin; dIncrement < dMax; dIncrement += dIntegrationInterval) {
				dResults[i] = 0;
				foreach (FuzzyVariableMembership m in dOutputVariables) {
					if (m.variable != null)
						dResults[i] = Math.Max(dResults[i], Math.Min(m.dMembership, m.variable.MembershipFunction(dIncrement)));
				}
				i++;
			}
			// Step 4: Find center of gravity. First integrate, then find midpoint of mass. This is the result.
			double dTotal = 0.0;
			for (i = 0; i < dResults.Length; i++)
				dTotal += dResults[i];
			dTotal /= 2.0;
			i = 0;
			for (double dIncrement = dMin; dIncrement < dMax; dIncrement += dIntegrationInterval) {
				dTotal -= dResults[i];
				if (dTotal < 0.0)
					return dIncrement;
				else
					i++;
			}
			return 0.0;  // Something went wrong.
		}

		#endregion

	}
}
